#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from enum import Enum
from cengal.parallel_execution.coroutines.coro_scheduler import *
from cengal.parallel_execution.coroutines.coro_tools.await_coro import *
from cengal.parallel_execution.coroutines.coro_standard_services.asyncio_loop import *
from cengal.parallel_execution.coroutines.coro_standard_services.loop_yield import CoroPriority
from cengal.parallel_execution.coroutines.coro_standard_services.put_coro import *
from cengal.parallel_execution.coroutines.coro_standard_services.timer_func_runner import *
from cengal.file_system.file_manager import path_relative_to_current_dir
from cengal.time_management.cpu_clock_cycles import perf_counter
from cengal.data_manipulation.serialization import *
from typing import Hashable, Tuple, List, Any, Dict, Callable, Type
from cengal.introspection.inspect import get_exception, entity_owning_module_importable_str, entity_owning_module_info_and_owning_path, entity_properties
from cengal.io.core.memory_management import IOCoreMemoryManagement
from cengal.parallel_execution.asyncio.efficient_streams import StreamManagerIOCoreMemoryManagement, TcpStreamManager, UdpStreamManager, StreamManagerAbstract
from cengal.code_flow_control.smart_values import ValueExistence
from cengal.io.named_connections.named_connections_manager import NamedConnectionsManager
from cengal.code_flow_control.args_manager import number_of_provided_args
from cengal.data_manipulation.serialization import Serializer, Serializers, best_serializer_for_standard_data
from cengal.code_flow_control.args_manager import find_arg_position_and_value, UnknownArgumentError
from cengal.data_generation.id_generator import IDGenerator, GeneratorType
from cengal.system import PLATFORM_NAME, PYTHON_VERSION
from importlib import import_module
import sys
import os
import asyncio

from .exceptions import *
from .commands import *


class LocalClassInfo:
    def __init__(self, local_id: Hashable, class_type: Type) -> None:
        self.class_type: Type = class_type
        self._class_name: str = self.class_type.__name__
        self._module_importable_str: str = entity_owning_module_importable_str(self.class_type)
        self._local_id: Hashable = local_id
    
    def __call__(self) -> Type:
        return {
            CommandDataFieldsDeclareServiceClass.local_id.value: self._local_id,
            CommandDataFieldsDeclareServiceClass.class_name.value: self._class_name,
            CommandDataFieldsDeclareServiceClass.module_importable_str.value: self._module_importable_str,
        }
    
    @property
    def class_name(self):
        return self._class_name
    
    @property
    def module_importable_str(self):
        return self._module_importable_str
    
    @property
    def local_id(self):
        return self._local_id


class RemoteClassInfo:
    def __init__(self, local_id: Hashable, class_name: str, module_importable_str: str) -> None:
        self.local_id: Hashable = local_id
        self.class_name: str = class_name
        self.module_importable_str: str = module_importable_str
        self.module = import_module(self.module_importable_str)
        self._class_type: Type = getattr(self.module, self.class_name)
    
    @classmethod
    def from_data(cls, data: Dict[Hashable, Any]) -> 'RemoteClassInfo':
        local_id: Hashable = data[CommandDataFieldsDeclareServiceClass.local_id.value]
        class_name: str = data[CommandDataFieldsDeclareServiceClass.class_name.value]
        module_importable_str: str = data[CommandDataFieldsDeclareServiceClass.module_importable_str.value]
        return cls(local_id, class_name, module_importable_str)
    
    def __call__(self) -> Type:
        return self._class_type
    
    @property
    def class_type(self):
        return self()



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from cengal.parallel_execution.asyncio.run_in_process_pool.versions.v_0.run_in_process_pool import *
import asyncio
from unittest import TestCase, main
from concurrent.futures import ProcessPoolExecutor


async def aworker(text: str) -> str:
    return text.upper()


async def worker(text: str) -> str:
    return text.upper()


class MyException(Exception):
    pass


async def aworker_ex(text: str) -> str:
    raise MyException('test')
    return text.upper()


async def worker_ex(text: str) -> str:
    raise MyException('test')
    return text.upper()


class TestProcessPool(TestCase):
    def test__multiprocess_async(self):
        async def coro(self):
            pp: ProcessPool = ProcessPool(ExecutorTypeSetup(ProcessPoolExecutor))
            pp.set_is_multiprocessing(True)
            result = await pp(aworker, 'test')
            self.assertEqual(result, 'TEST')
        
        asyncio.run(coro(self))

    def test__singleprocess_async(self):
        async def coro(self):
            pp: ProcessPool = ProcessPool(ExecutorTypeSetup(ProcessPoolExecutor))
            pp.set_is_multiprocessing(False)
            result = await pp(aworker, 'test')
            self.assertEqual(result, 'TEST')
        
        asyncio.run(coro(self))

    def test__multiprocess_async_ex(self):
        async def coro(self):
            pp: ProcessPool = ProcessPool(ExecutorTypeSetup(ProcessPoolExecutor))
            pp.set_is_multiprocessing(True)
            with self.assertRaises(MyException) as cm:
                result = await pp(aworker_ex, 'test')
            
            self.assertEqual(str(cm.exception), 'test')
        
        asyncio.run(coro(self))

    def test__multiprocess_sync_ex(self):
        async def coro(self):
            pp: ProcessPool = ProcessPool(ExecutorTypeSetup(ProcessPoolExecutor))
            pp.set_is_multiprocessing(True)
            with self.assertRaises(MyException) as cm:
                result = await pp(worker_ex, 'test')
            
            self.assertEqual(str(cm.exception), 'test')
        
        asyncio.run(coro(self))


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from cengal.text_processing.encoding_detection import detect_and_decode
from cengal.text_processing.brackets_processing import find_text_in_brackets, BracketPair
from cengal.text_processing.text_processing import Text, replace_text, replace_slice, DEFAULT_ENCODING
from typing import List, Tuple, Optional, Callable


def patch_text(
    text: Text, 
    patch: List[Tuple[BracketPair, Text]], 
    count: int = 1, 
    encoding: Optional[str] = DEFAULT_ENCODING, 
    normalizer: Optional[Callable] = None
    ) -> Text:
    iter_num: int = 0
    sub_iter_num: int = -1
    while (0 != sub_iter_num) and ((iter_num < count) if (0 <= count) else True):
        iter_num += 1
        sub_iter_num = 0
        for brackets_pair, after in patch:
            old_text_slice = find_text_in_brackets(text, brackets_pair, encoding=encoding, normalizer=normalizer)
            if old_text_slice is None:
                continue

            text, _ = replace_slice(text, old_text_slice, after, encoding=encoding, normalizer=normalizer)
            sub_iter_num += 1

    return text



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


__all__ = ['prepare_loop', 'restore_loop', 'DerivedFromProactorEventLoop', 'DerivedFromSelectorEventLoop', 'DerivedFromUVLoop']


from cengal.parallel_execution.coroutines.coro_scheduler import CoroSchedulerType, Interface, Coro, AnyWorker, current_interface, current_coro_scheduler, cs_coro, cs_acoro
from cengal.data_manipulation.conversion.reinterpret_cast import reinterpret_cast
from cengal.data_manipulation.conversion.reinterpret_cast_management.manager import BaseAutoDerivedObjWrapper
from asyncio import SelectorEventLoop, AbstractEventLoop
_proactor_present = False
try:
    from asyncio import ProactorEventLoop
    _proactor_present = True
except ImportError:
    pass

from uvloop import Loop as UVLoop
from typing import Type, Tuple, Dict, Callable, Any


def call_soon(self, callback, *args, **kwargs):
    from .asyncio_loop import AsyncioLoop
    cs: CoroSchedulerType = self._cs
    service: AsyncioLoop = cs.get_service_instance(AsyncioLoop)
    service.register_new_asyncio_request()
    return SelectorEventLoop.call_soon(self, callback, *args, **kwargs)

def call_later(self, delay, callback, *args, **kwargs):
    from .asyncio_loop import AsyncioLoop
    cs: CoroSchedulerType = self._cs
    service: AsyncioLoop = cs.get_service_instance(AsyncioLoop)
    service.register_new_asyncio_request()
    return SelectorEventLoop.call_later(self, delay, callback, *args, **kwargs)

def call_at(self, when, callback, *args, **kwargs):
    from .asyncio_loop import AsyncioLoop
    cs: CoroSchedulerType = self._cs
    service: AsyncioLoop = cs.get_service_instance(AsyncioLoop)
    service.register_new_asyncio_request()
    return SelectorEventLoop.call_at(self, when, callback, *args, **kwargs)

# Method scheduling a coroutine object: create a task.

def create_task(self, coro, **kwargs):
    from .asyncio_loop import AsyncioLoop
    cs: CoroSchedulerType = self._cs
    service: AsyncioLoop = cs.get_service_instance(AsyncioLoop)
    service.register_new_asyncio_request()
    return SelectorEventLoop.create_task(self, coro, **kwargs)

# Methods for interacting with threads.

def call_soon_threadsafe(self, callback, *args, **kwargs):
    from .asyncio_loop import AsyncioLoop
    cs: CoroSchedulerType = self._cs
    service: AsyncioLoop = cs.get_service_instance(AsyncioLoop)
    service.register_new_asyncio_request()
    return SelectorEventLoop.call_soon_threadsafe(self, callback, *args, **kwargs)

def run_in_executor(self, executor, func, *args):
    from .asyncio_loop import AsyncioLoop
    cs: CoroSchedulerType = self._cs
    service: AsyncioLoop = cs.get_service_instance(AsyncioLoop)
    service.register_new_asyncio_request()
    return SelectorEventLoop.run_in_executor(self, executor, func, *args)

def add_reader(self, fd, callback, *args):
    from .asyncio_loop import AsyncioLoop
    cs: CoroSchedulerType = self._cs
    service: AsyncioLoop = cs.get_service_instance(AsyncioLoop)
    service.register_new_asyncio_request()
    return SelectorEventLoop.add_reader(self, fd, callback, *args)

def add_writer(self, fd, callback, *args):
    from .asyncio_loop import AsyncioLoop
    cs: CoroSchedulerType = self._cs
    service: AsyncioLoop = cs.get_service_instance(AsyncioLoop)
    service.register_new_asyncio_request()
    return SelectorEventLoop.add_writer(self, fd, callback, *args)

# Signal handling.

def add_signal_handler(self, sig, callback, *args):
    from .asyncio_loop import AsyncioLoop
    cs: CoroSchedulerType = self._cs
    service: AsyncioLoop = cs.get_service_instance(AsyncioLoop)
    service.register_new_asyncio_request()
    return SelectorEventLoop.add_signal_handler(self, sig, callback, *args)


class LoopWrapper(BaseAutoDerivedObjWrapper):
    def wrapping_required(self, obj: Any, base_type: Type, fields: Tuple[str], planned_type_name: str) -> bool:
        return issubclass(base_type, AbstractEventLoop)
    
    def methods(self, obj: Any, base_type: Type, fields: Tuple[str]) -> Dict[str, Callable]:
        return {
            'call_soon': call_soon,
            'call_later': call_later,
            'call_at': call_at,
            'create_task': create_task,
            'call_soon_threadsafe': call_soon_threadsafe,
            'run_in_executor': run_in_executor,
            'add_reader': add_reader,
            'add_writer': add_writer,
            'add_signal_handler': add_signal_handler,
        }


_lw: LoopWrapper = LoopWrapper()


def prepare_loop(loop: AbstractEventLoop) -> Type:
    if _proactor_present and type(loop) is ProactorEventLoop:
        reinterpret_cast(DerivedFromProactorEventLoop, loop)
        loop._cs = current_coro_scheduler()
        return ProactorEventLoop
    elif type(loop) is SelectorEventLoop:
        reinterpret_cast(DerivedFromSelectorEventLoop, loop)
        loop._cs = current_coro_scheduler()
        return SelectorEventLoop
    elif type(loop) is UVLoop:
        reinterpret_cast(DerivedFromUVLoop, loop)
        loop._cs = current_coro_scheduler()
        return UVLoop
    elif isinstance(loop, AbstractEventLoop):
        original_type: Type = type(loop)
        _lw(loop)
        loop._cs = current_coro_scheduler()
        return original_type
    else:
        raise TypeError('Unknown loop type: {}'.format(type(loop)))


def restore_loop(loop: AbstractEventLoop, original_type: Type) -> None:
    reinterpret_cast(original_type, loop)


if _proactor_present:
    class DerivedFromProactorEventLoop(ProactorEventLoop):
        def call_soon(self, callback, *args, **kwargs):
            from .asyncio_loop import AsyncioLoop
            cs: CoroSchedulerType = self._cs
            service: AsyncioLoop = cs.get_service_instance(AsyncioLoop)
            service.register_new_asyncio_request()
            return ProactorEventLoop.call_soon(self, callback, *args, **kwargs)

        def call_later(self, delay, callback, *args, **kwargs):
            from .asyncio_loop import AsyncioLoop
            cs: CoroSchedulerType = self._cs
            service: AsyncioLoop = cs.get_service_instance(AsyncioLoop)
            service.register_new_asyncio_request()
            return ProactorEventLoop.call_later(self, delay, callback, *args, **kwargs)

        def call_at(self, when, callback, *args, **kwargs):
            from .asyncio_loop import AsyncioLoop
            cs: CoroSchedulerType = self._cs
            service: AsyncioLoop = cs.get_service_instance(AsyncioLoop)
            service.register_new_asyncio_request()
            return ProactorEventLoop.call_at(self, when, callback, *args, **kwargs)

        # Method scheduling a coroutine object: create a task.

        def create_task(self, coro, **kwargs):
            from .asyncio_loop import AsyncioLoop
            cs: CoroSchedulerType = self._cs
            service: AsyncioLoop = cs.get_service_instance(AsyncioLoop)
            service.register_new_asyncio_request()
            return ProactorEventLoop.create_task(self, coro, **kwargs)

        # Methods for interacting with threads.

        def call_soon_threadsafe(self, callback, *args, **kwargs):
            from .asyncio_loop import AsyncioLoop
            cs: CoroSchedulerType = self._cs
            service: AsyncioLoop = cs.get_service_instance(AsyncioLoop)
            service.register_new_asyncio_request()
            return ProactorEventLoop.call_soon_threadsafe(self, callback, *args, **kwargs)

        def run_in_executor(self, executor, func, *args):
            from .asyncio_loop import AsyncioLoop
            cs: CoroSchedulerType = self._cs
            service: AsyncioLoop = cs.get_service_instance(AsyncioLoop)
            service.register_new_asyncio_request()
            return ProactorEventLoop.run_in_executor(self, executor, func, *args)

        def add_reader(self, fd, callback, *args):
            from .asyncio_loop import AsyncioLoop
            cs: CoroSchedulerType = self._cs
            service: AsyncioLoop = cs.get_service_instance(AsyncioLoop)
            service.register_new_asyncio_request()
            return ProactorEventLoop.add_reader(self, fd, callback, *args)

        def add_writer(self, fd, callback, *args):
            from .asyncio_loop import AsyncioLoop
            cs: CoroSchedulerType = self._cs
            service: AsyncioLoop = cs.get_service_instance(AsyncioLoop)
            service.register_new_asyncio_request()
            return ProactorEventLoop.add_writer(self, fd, callback, *args)

        # Signal handling.

        def add_signal_handler(self, sig, callback, *args):
            from .asyncio_loop import AsyncioLoop
            cs: CoroSchedulerType = self._cs
            service: AsyncioLoop = cs.get_service_instance(AsyncioLoop)
            service.register_new_asyncio_request()
            return ProactorEventLoop.add_signal_handler(self, sig, callback, *args)


class DerivedFromSelectorEventLoop(SelectorEventLoop):
    def call_soon(self, callback, *args, **kwargs):
        from .asyncio_loop import AsyncioLoop
        cs: CoroSchedulerType = self._cs
        service: AsyncioLoop = cs.get_service_instance(AsyncioLoop)
        service.register_new_asyncio_request()
        return SelectorEventLoop.call_soon(self, callback, *args, **kwargs)

    def call_later(self, delay, callback, *args, **kwargs):
        from .asyncio_loop import AsyncioLoop
        cs: CoroSchedulerType = self._cs
        service: AsyncioLoop = cs.get_service_instance(AsyncioLoop)
        service.register_new_asyncio_request()
        return SelectorEventLoop.call_later(self, delay, callback, *args, **kwargs)

    def call_at(self, when, callback, *args, **kwargs):
        from .asyncio_loop import AsyncioLoop
        cs: CoroSchedulerType = self._cs
        service: AsyncioLoop = cs.get_service_instance(AsyncioLoop)
        service.register_new_asyncio_request()
        return SelectorEventLoop.call_at(self, when, callback, *args, **kwargs)

    # Method scheduling a coroutine object: create a task.

    def create_task(self, coro, **kwargs):
        from .asyncio_loop import AsyncioLoop
        cs: CoroSchedulerType = self._cs
        service: AsyncioLoop = cs.get_service_instance(AsyncioLoop)
        service.register_new_asyncio_request()
        return SelectorEventLoop.create_task(self, coro, **kwargs)

    # Methods for interacting with threads.

    def call_soon_threadsafe(self, callback, *args, **kwargs):
        from .asyncio_loop import AsyncioLoop
        cs: CoroSchedulerType = self._cs
        service: AsyncioLoop = cs.get_service_instance(AsyncioLoop)
        service.register_new_asyncio_request()
        return SelectorEventLoop.call_soon_threadsafe(self, callback, *args, **kwargs)

    def run_in_executor(self, executor, func, *args):
        from .asyncio_loop import AsyncioLoop
        cs: CoroSchedulerType = self._cs
        service: AsyncioLoop = cs.get_service_instance(AsyncioLoop)
        service.register_new_asyncio_request()
        return SelectorEventLoop.run_in_executor(self, executor, func, *args)

    def add_reader(self, fd, callback, *args):
        from .asyncio_loop import AsyncioLoop
        cs: CoroSchedulerType = self._cs
        service: AsyncioLoop = cs.get_service_instance(AsyncioLoop)
        service.register_new_asyncio_request()
        return SelectorEventLoop.add_reader(self, fd, callback, *args)

    def add_writer(self, fd, callback, *args):
        from .asyncio_loop import AsyncioLoop
        cs: CoroSchedulerType = self._cs
        service: AsyncioLoop = cs.get_service_instance(AsyncioLoop)
        service.register_new_asyncio_request()
        return SelectorEventLoop.add_writer(self, fd, callback, *args)

    # Signal handling.

    def add_signal_handler(self, sig, callback, *args):
        from .asyncio_loop import AsyncioLoop
        cs: CoroSchedulerType = self._cs
        service: AsyncioLoop = cs.get_service_instance(AsyncioLoop)
        service.register_new_asyncio_request()
        return SelectorEventLoop.add_signal_handler(self, sig, callback, *args)


class DerivedFromUVLoop(UVLoop):
    def call_soon(self, callback, *args, context=None):
        from .asyncio_loop import AsyncioLoop
        cs: CoroSchedulerType = self._cs
        service: AsyncioLoop = cs.get_service_instance(AsyncioLoop)
        service.register_new_asyncio_request()
        return UVLoop.call_soon(self, callback, *args, context=context)

    def call_later(self, delay, callback, *args, context=None):
        from .asyncio_loop import AsyncioLoop
        cs: CoroSchedulerType = self._cs
        service: AsyncioLoop = cs.get_service_instance(AsyncioLoop)
        service.register_new_asyncio_request()
        return UVLoop.call_later(self, delay, callback, *args, context=context)

    def call_at(self, when, callback, *args, context=None):
        from .asyncio_loop import AsyncioLoop
        cs: CoroSchedulerType = self._cs
        service: AsyncioLoop = cs.get_service_instance(AsyncioLoop)
        service.register_new_asyncio_request()
        return UVLoop.call_at(self, when, callback, *args, context=context)

    # Method scheduling a coroutine object: create a task.

    def create_task(self, coro, *, name=None, context=None):
        from .asyncio_loop import AsyncioLoop
        cs: CoroSchedulerType = self._cs
        service: AsyncioLoop = cs.get_service_instance(AsyncioLoop)
        service.register_new_asyncio_request()
        return UVLoop.create_task(self, coro, name=name, context=context)

    # Methods for interacting with threads.

    def call_soon_threadsafe(self, callback, *args, context=None):
        from .asyncio_loop import AsyncioLoop
        cs: CoroSchedulerType = self._cs
        service: AsyncioLoop = cs.get_service_instance(AsyncioLoop)
        service.register_new_asyncio_request()
        return UVLoop.call_soon_threadsafe(self, callback, *args, context=context)

    def run_in_executor(self, executor, func, *args):
        from .asyncio_loop import AsyncioLoop
        cs: CoroSchedulerType = self._cs
        service: AsyncioLoop = cs.get_service_instance(AsyncioLoop)
        service.register_new_asyncio_request()
        return UVLoop.run_in_executor(self, executor, func, *args)

    def add_reader(self, fd, callback, *args):
        from .asyncio_loop import AsyncioLoop
        cs: CoroSchedulerType = self._cs
        service: AsyncioLoop = cs.get_service_instance(AsyncioLoop)
        service.register_new_asyncio_request()
        return UVLoop.add_reader(self, fd, callback, *args)

    def add_writer(self, fd, callback, *args):
        from .asyncio_loop import AsyncioLoop
        cs: CoroSchedulerType = self._cs
        service: AsyncioLoop = cs.get_service_instance(AsyncioLoop)
        service.register_new_asyncio_request()
        return UVLoop.add_writer(self, fd, callback, *args)

    # Signal handling.

    def add_signal_handler(self, sig, callback, *args):
        from .asyncio_loop import AsyncioLoop
        cs: CoroSchedulerType = self._cs
        service: AsyncioLoop = cs.get_service_instance(AsyncioLoop)
        service.register_new_asyncio_request()
        return UVLoop.add_signal_handler(self, sig, callback, *args)



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from cengal.data_generation.id_generator import IDGenerator
from cengal.data_containers.fast_fifo import FIFODequeWithLengthControl
from cengal.code_flow_control.smart_values import ValueCache, ValueExistence
from weakref import ref, ReferenceType
from typing import Optional, Set, Dict, List, Tuple, Callable, Awaitable, Sequence, Type


# class NamedConnectionsTransport:
#     def disconnect(self):
#         raise NotImplementedError
    
#     def disconnected(self):
#         ...
    
#     def eof(self) -> bool:
#         ...


class NamedConnectionsClient:
    def __init__(self, server_id: bytes, external_data_full_size: ValueExistence) -> None:
        self.id = None
        self.server_id: bytes = server_id
        self.server_ref: Optional[ReferenceType['NamedConnectionsServer']] = None
        self.input_from_client: FIFODequeWithLengthControl = FIFODequeWithLengthControl(external_data_full_size)
        self.output_to_client: FIFODequeWithLengthControl = FIFODequeWithLengthControl(external_data_full_size)
    
    def callback__bind(self, server: 'NamedConnectionsServer'):
        """Server can emit this callback

        Args:
            server (NamedConnectionsServer): _description_
        """
        self.server_ref = ref(server)
    
    def callback__unbind(self):
        """Server can emit this callback
        """
        self.server_ref = None

    def callback__data_to_client_added(self):
        """Server can emit this callback

        Raises:
            NotImplementedError: _description_
        """
        raise NotImplementedError
    
    def callback__server_ready_for_data(self):
        """Server can emit this callback

        Raises:
            NotImplementedError: _description_
        """
        raise NotImplementedError

    def callback__is_client_ready_for_data(self) -> bool:
        """Server can emit this callback

        Raises:
            NotImplementedError: _description_

        Returns:
            bool: _description_
        """
        raise NotImplementedError
    
    def callback__stop(self):
        """Server can emit this callback. No additional data from the server will be provided. No read from the server side will be made.

        Raises:
            NotImplementedError: _description_
        """
        raise NotImplementedError
    
    def data_to_server_added(self):
        """Client can call this method
        """
        server = self.server_ref()
        if server:
            server.callback__data_to_server_added(self)
    
    def is_server_ready_for_data(self) -> bool:
        """Client can call this method

        Returns:
            bool: _description_
        """
        server = self.server_ref()
        if server:
            return server.callback__is_server_ready_for_data(self)
        else:
            return False

    def client_ready_for_data(self):
        """Client can call this method
        """
        server = self.server_ref()
        if server:
            server.callback__client_ready_for_data(self)
    
    def stop(self):
        """Client can call this method. Server should not call it. Server should call NamedConnectionsClient.callback__stop() when wants to close connection
        """
        server = self.server_ref()
        if server:
            server.callback__client_stopped(self)



class NamedConnectionsServer:
    # def __init__(self, server_id: bytes, named_connections_manager: Optional['NamedConnectionsManager'] = None) -> None:
    #     self.id: bytes = server_id
    #     if named_connections_manager:
    #         self.named_connections_manager: ReferenceType['NamedConnectionsManager'] = ref(named_connections_manager)
    #     else:
    #         self.named_connections_manager = None
        
    #     self.bind_clients: Dict[int, NamedConnectionsClient] = dict()

    def __init__(self, server_id: bytes, named_connections_manager: 'NamedConnectionsManager') -> None:
        self.id: bytes = server_id
        self.named_connections_manager: ReferenceType['NamedConnectionsManager'] = ref(named_connections_manager)
        self.bind_clients: Dict[int, NamedConnectionsClient] = dict()

    def bind(self, client: NamedConnectionsClient):
        self.bind_clients[client.id] = client
        client.callback__bind(self)

    def unbind(self, client: NamedConnectionsClient):
        try:
            del self.bind_clients[client.id]
        except KeyError:
            pass
        finally:
            client.callback__unbind()
    
    def callback__data_to_server_added(self, client: NamedConnectionsClient):
        raise NotImplementedError
    
    def callback__is_server_ready_for_data(self, client: NamedConnectionsClient) -> bool:
        raise NotImplementedError

    def callback__client_ready_for_data(self, client: NamedConnectionsClient):
        raise NotImplementedError
    
    def callback__client_stopped(self, client: NamedConnectionsClient):
        """Client can emit this callback. No additional data from the client will be provided. No read from the client side will be made.

        Raises:
            NotImplementedError: _description_
        """
        raise NotImplementedError


class NamedConnectionsManager:
    def __init__(self, external_data_full_size: ValueExistence) -> None:
        self.external_data_full_size: ValueExistence = external_data_full_size
        self.gen_client_id: IDGenerator = IDGenerator()
        self.unbind_clients: Dict[int, NamedConnectionsClient] = dict()
        self.unbind_clients_per_server: Dict[str, Set[NamedConnectionsClient]] = dict()
        self.bind_clients: Dict[int, NamedConnectionsClient] = dict()
        self.servers: Dict[bytes, NamedConnectionsServer] = dict()
    
    def create_client(self, client_type: Type[NamedConnectionsClient], server_id: bytes) -> NamedConnectionsClient:
        client: NamedConnectionsClient = client_type(server_id, self.external_data_full_size)
        client_id = self.gen_client_id()
        client.id = client_id
        if server_id in self.servers:
            self.bind_clients[client_id] = client
            self.servers[server_id].bind(client)
        else:
            self._unbind_client_impl(client)
        
        return client_id
    
    def register_client(self, client: NamedConnectionsClient) -> int:
        server_id: bytes = client.server_id
        client_id = self.gen_client_id()
        client.id = client_id
        if server_id in self.servers:
            self.bind_clients[client_id] = client
            self.servers[server_id].bind(client)
        else:
            self.unbind_clients[client_id] = client
            if server_id not in self.unbind_clients_per_server:
                self.unbind_clients_per_server[server_id] = set()
            
            self.unbind_clients_per_server[server_id].add(client)
        
        return client_id

    def register_server(self, server: NamedConnectionsServer):
        server_id: bytes = server.id
        if server_id in self.servers:
            raise RuntimeError('NamedConnectionsServer already registered')
        
        self.servers[server_id] = server
        if server_id in self.unbind_clients_per_server:
            unbind_clients = self.unbind_clients_per_server[server_id]
            del self.unbind_clients_per_server[server_id]
            for client in unbind_clients:
                del self.unbind_clients[client.id]
                self.bind_clients[client.id] = client
                server.bind(client)

    def unregister_server(self, server: NamedConnectionsServer):
        server_clients = server.bind_clients.items()
        for client_id, client in server_clients:
            self.unbind_client(client)
        
        try:
            del self.servers[server.id]
        except KeyError:
            pass

    def unregister_client(self, client: NamedConnectionsClient):
        server = client.server_ref()
        if server:
            server.unbind(client)
        
        client_id = client.id
        self.bind_clients.pop(client_id, None)
        self.unbind_clients.pop(client_id, None)
        try:
            self.unbind_clients_per_server[client.server_id].discard(client)
        except KeyError:
            pass

    def unbind_client(self, client: NamedConnectionsClient):
        server = client.server_ref()
        if server:
            server.unbind(client)

        self._unbind_client_impl(client)

    def _unbind_client_impl(self, client: NamedConnectionsClient):
        client_id = client.id
        server_id = client.server_id
        self.bind_clients.pop(client_id, None)
        self.unbind_clients[client_id] = client
        if server_id not in self.unbind_clients_per_server:
            self.unbind_clients_per_server[server_id] = set()
        
        self.unbind_clients_per_server[server_id].add(client)

    def rebind_client(self, client: NamedConnectionsClient, server_id: bytes):
        """Force rebind client to another server
        """
        self.servers[client.server_id].unbind(client)
        client.server_id = server_id
        self.servers[server_id].bind(client)
    
    # def put_client_message_to_server_queue(self, message):
    #     ...
    
    # def put_server_message_to_client_queue(self, message):
    #     ...
    
    # def client_to_server_queue_len(self):
    #     ...
    
    # def server_to_client_queue_len(self):
    #     ...



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


__all__ = ['ShutdownOnKeyboardInterrupt']


from cengal.parallel_execution.coroutines.coro_scheduler import *
import signal
from typing import Tuple


class ShutdownOnKeyboardInterrupt(TypedService[None]):
    def __init__(self, loop: CoroSchedulerType):
        super().__init__(loop)
        self.handler_set: bool = False
        self.previous_handler = None
        self.keyboard_interrupt_emited: bool = False

    def single_task_registration_or_immediate_processing(
            self, *args, **kwargs) -> Tuple[bool, None, None]:
        if not self.handler_set:
            self.previous_handler = signal.signal(signal.SIGINT, self.keyboard_interrupt_handler)
            self.handler_set = True
        
        return True, None, None

    def full_processing_iteration(self):
        if self.keyboard_interrupt_emited:
            signal.signal(signal.SIGINT, self.previous_handler)
            self.handler_set = False
            if callable(self.previous_handler):
                def on_cs_destroy_handler():
                    self.previous_handler(signal.SIGINT, None)
                
                self._loop.on_destroyed_handlers.add(on_cs_destroy_handler)

            raise CoroSchedulerDestroyRequestedException
        
        self.make_dead()

    def in_work(self) -> bool:
        result: bool = self.keyboard_interrupt_emited
        return self.thrifty_in_work(result)
    
    def keyboard_interrupt_handler(self, sig, frame):
        self.keyboard_interrupt_emited = True
        self.make_live()
    
    def destroy(self):
        if self.handler_set:
            signal.signal(signal.SIGINT, self.previous_handler)
            self.handler_set = False



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from os import rename, remove
from os.path import sep as os_sep, dirname, exists, isfile
from distutils.dir_util import copy_tree
from distutils.file_util import copy_file
from cengal.file_system.win_fs.global_install_uninstall import win_fs
from cengal.file_system.path_manager import path_relative_to_src, RelativePath
from cengal.file_system.file_manager import file_exists
from cengal.text_processing.text_processing import removeprefix
from cengal.file_system.file_patch.simple import patch_text_file
import os
from typing import Set


create_env_variants: Set[str] = {
    'create_env.sh',
    'create_env.cmd',
    'command/_common.sh',
    'command/_common.cmd',
}


def create(src_dir: str, module_submodule_path: str, sep: str = '/'):
    with win_fs():
        src_dir_path = RelativePath(src_dir)
        template_module_dir = src_dir_path('_template_module')
        template_submodule_dir = src_dir_path('_template_module/_template_submodule')
        required_path_input = module_submodule_path
        required_path_input = required_path_input.replace(os_sep, sep)
        required_path_input = required_path_input.strip()
        required_path_input = required_path_input.strip(sep)
        # if sep not in required_path_input:
        #     print('Wrong path: both module and submodule are mandatory')
        #     return
        
        required_path_input_parts = required_path_input.split(sep)
        submodule_name = required_path_input_parts[-1]
        submodule_path = src_dir_path(required_path_input)
        module_path = RelativePath(submodule_path)('..')
        copy_tree(template_submodule_dir, submodule_path, update=True)

        template_submodule_py_path = RelativePath(submodule_path)('versions/v_0/_template_submodule.py')
        template_submodule_py_new_path = RelativePath(submodule_path)(f'versions/v_0/{submodule_name}.py')
        if exists(template_submodule_py_new_path) and isfile(template_submodule_py_new_path):
            remove(template_submodule_py_path)
        else:
            rename(template_submodule_py_path, template_submodule_py_new_path)

        init_py_path = RelativePath(submodule_path)('versions/v_0/__init__.py')
        patch_text_file(init_py_path, [('_template_submodule', submodule_name)])

        for index in range(1, 1 + len(required_path_input_parts)):
            subpath: str = sep.join(required_path_input_parts[:index])
            subpath_init: str = src_dir_path(subpath + sep + '__init__.py')
            if not file_exists(subpath_init):
                copy_file(src_dir_path('_template_module/__init__.py'), subpath_init, verbose=0)

        # create_env_sh_path = RelativePath(submodule_path)('versions/v_0/development/create_env.sh')
        # development_dir_path = dirname(create_env_sh_path)
        development_dir_path = RelativePath(submodule_path)('versions/v_0/development')
        remaining_path = removeprefix(development_dir_path, src_dir)
        sep_num = remaining_path.count(sep)
        cd_2_parent_num = 1 + sep_num  # since we need dir parent to the `src_dir`
        new_relativeness = '/..' * cd_2_parent_num
        new_relativeness_win32 = '\\..' * cd_2_parent_num

        development_dir_path_rel: RelativePath = RelativePath(development_dir_path)
        for create_env_variant in create_env_variants:
            create_env_sh_path = development_dir_path_rel(create_env_variant)
            if not file_exists(create_env_sh_path):
                continue
            
            patch_text_file(create_env_sh_path, [
                ('$WORKDIR/../../../../../..`', f'$WORKDIR{new_relativeness}`'),
                ('$CURRENT_PROJECT_DIR_PATH/../../../../../.."', f'$CURRENT_PROJECT_DIR_PATH{new_relativeness}"'),
                ('%CURRENT_PROJECT_DIR_PATH%\..\..\..\..\..\.."', f'%CURRENT_PROJECT_DIR_PATH%{new_relativeness_win32}"'),
            ])



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ['CallHistoryReapplier', 'CoroPriority']


from typing import Dict, Tuple, Hashable, Any
from cengal.parallel_execution.coroutines.coro_standard_services.loop_yield import *


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


class CallHistoryReapplier:
    def __init__(self, priority: CoroPriority=CoroPriority.normal):
        self.history: Dict = dict()
        self.priority = priority
    
    def call_impl(self, *args, **kwargs):
        raise NotImplementedError
    
    def args_to_key_value(self, *args, **kwargs) -> Tuple[Hashable, Any]:
        raise NotImplementedError
    
    def key_value_to_args(self, key: Hashable, value: Any) -> Tuple[Tuple, Dict]:
        raise NotImplementedError
    
    def __call__(self, *args, **kwargs):
        key, value = self.args_to_key_value(*args, **kwargs)
        self.history[key] = value
        self.call_impl(*args, **kwargs)
    
    def reapply(self):
        ly = gly(self.priority)
        new_history = dict()
        while self.history:
            history_buff = self.history
            self.history = type(self.history)()
            for key, value in history_buff.items():
                args, kwargs = self.key_value_to_args(key, value)
                new_history[key] = value
                self.call_impl(*args, **kwargs)
                ly()
        
        # in order to not lose call history made in intermediate of the reapplying process:
        # self.history = new_history
        new_history.update(self.history)
        self.history = new_history

    def destroy(self):
        self.history = type(self.history)()



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['compiler_type', 'compiler_name', 'compiler_string', 'compiler_string_escaped']


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from distutils import sysconfig
from cengal.os.execute import escape_text


compiler_name_raw: str = sysconfig.get_config_var("CC")
if compiler_name_raw is None:
    if sysconfig.get_config_var("VSINSTALLDIR") is None:
        compiler_name_raw = str()
    else:
        compiler_name_raw = 'msvc'

compiler_name: str = compiler_name_raw.split()[0] if compiler_name_raw else str()
compiler_string: str = compiler_name_raw
compiler_string_escaped: str = escape_text(compiler_name_raw)

if "gcc" in compiler_name_raw.lower():
    compiler_type: str = 'gcc'
elif "msvc" in compiler_name_raw.lower():
    compiler_type = 'msvc'
elif "clang" in compiler_name_raw.lower():
    compiler_type = 'clang'
elif "icc" in compiler_name_raw.lower():
    compiler_type = 'icc'
elif "llvm" in compiler_name_raw.lower():
    compiler_type = 'llvm'
elif "intel" in compiler_name_raw.lower():
    compiler_type = 'intel'
elif "arm" in compiler_name_raw.lower():
    compiler_type = 'arm'
elif "mingw" in compiler_name_raw.lower():
    compiler_type = 'mingw'
else:
    compiler_type = 'unknown'


#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# __all__ = []
__all__ = [
    'WaitType',
    'TasksListIsEmptyError',
    'IsNotDoneYetError',
    'NormalOnDoneHandler',
    'NormalInfo',
    'FastestOnDoneHandler',
    'FastestInfo',
    'FastestSuccessfulOnDoneHandler',
    'FastestSuccessfulInfo',
    'FastestExceptionOnDoneHandler',
    'FastestExceptionInfo',
    'FastestCustomOnDoneHandler',
    'FastestCustomInfo',
    'AtomicOnDoneHandler',
    'AtomicInfo',
    'AtomicCustomOnDoneHandler',
    'success_result_criteria_handler',
    'exception_result_criteria_handler',
    'same_first_failed_result_formatter',
    'none_first_failed_result_formatter',
    'AtomicCustomInfo',
    'GracefulTerminationSettings',
    'TerminationReason',
    'atask_graceful_destroyer',
    'aterminate_tasks_explicit',
    'aterminate_tasks_implicit',
    'aterminate_tasks',
    'aterminate_tasks_im',
    'terminate_tasks_explicit',
    'terminate_tasks_implicit',
    'terminate_tasks',
    'terminate_tasks_im',
    'await_tasks_explicit',
    'await_tasks_implicit',
    'await_tasks',
    'await_tasks_im',
    'wait_tasks_explicit',
    'wait_tasks_implicit',
    'wait_tasks',
    'wait_tasks_im',
    'await_tasks_fastest_explicit',
    'await_tasks_fastest_implicit',
    'await_tasks_fastest',
    'await_tasks_fastest_im',
    'wait_tasks_fastest_explicit',
    'wait_tasks_fastest_implicit',
    'wait_tasks_fastest',
    'wait_tasks_fastest_im',
    'await_tasks_fastest_successful_explicit',
    'await_tasks_fastest_successful_implicit',
    'await_tasks_fastest_successful',
    'await_tasks_fastest_successful_im',
    'wait_tasks_fastest_successful_explicit',
    'wait_tasks_fastest_successful_implicit',
    'wait_tasks_fastest_successful',
    'wait_tasks_fastest_successful_im',
    'await_tasks_fastest_exception_explicit',
    'await_tasks_fastest_exception_implicit',
    'await_tasks_fastest_exception',
    'await_tasks_fastest_exception_im',
    'wait_tasks_fastest_exception_explicit',
    'wait_tasks_fastest_exception_implicit',
    'wait_tasks_fastest_exception',
    'wait_tasks_fastest_exception_im',
    'await_tasks_fastest_custom_explicit',
    'await_tasks_fastest_custom_implicit',
    'await_tasks_fastest_custom',
    'await_tasks_fastest_custom_im',
    'wait_tasks_fastest_custom_explicit',
    'wait_tasks_fastest_custom_implicit',
    'wait_tasks_fastest_custom',
    'wait_tasks_fastest_custom_im',
    'await_tasks_atomic_explicit',
    'await_tasks_atomic_implicit',
    'await_tasks_atomic',
    'await_tasks_atomic_im',
    'wait_tasks_atomic_explicit',
    'wait_tasks_atomic_implicit',
    'wait_tasks_atomic',
    'wait_tasks_atomic_im',
    'await_tasks_atomic_custom_explicit',
    'await_tasks_atomic_custom_implicit',
    'await_tasks_atomic_custom',
    'await_tasks_atomic_custom_im',
    'wait_tasks_atomic_custom_explicit',
    'wait_tasks_atomic_custom_implicit',
    'wait_tasks_atomic_custom',
    'wait_tasks_atomic_custom_im',
    'wait_task_explicit',
    'wait_task_implicit',
    'wait_task',
    'wait_task_im',
    'await_task_explicit',
    'await_task_implicit',
    'await_task',
    'await_task_im',
]


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from cengal.parallel_execution.coroutines.coro_scheduler import Interface, current_interface, CoroID, get_interface_and_loop_with_explicit_loop
from cengal.parallel_execution.coroutines.coro_standard_services.put_coro import Task
from cengal.parallel_execution.coroutines.coro_standard_services.put_coro_list import PutCoroList, PSCP
from cengal.parallel_execution.coroutines.coro_standard_services.run_coro import RunCoro
from cengal.parallel_execution.coroutines.coro_standard_services.kill_coro_list import KillCoroList, KillSingleCoroParams
from cengal.parallel_execution.coroutines.coro_standard_services.wait_coro import TimeoutError, SubTimeoutError
from cengal.parallel_execution.coroutines.coro_standard_services.timer_func_runner import timer_func_run_on
from cengal.parallel_execution.coroutines.coro_tools.coro_flow_control import agraceful_coro_destroyer
from cengal.parallel_execution.coroutines.coro_standard_services.async_event_bus import AsyncEventBusRequest, AsyncEventBus
from cengal.parallel_execution.coroutines.coro_standard_services.loop_yield import CoroPriority
from cengal.parallel_execution.coroutines.coro_standard_services_internal_lib.service_with_a_direct_request import put_request_to_service

from uuid import uuid4
from enum import IntEnum
from typing import Optional, List, Union, Any, Type, Tuple, Dict, Callable


class WaitType(IntEnum):
    normal = 0
    fastest = 1
    fastest_successful = 2
    fastest_exception = 3
    fastest_custom = 4
    atomic = 5
    atomic_custom = 5


class TasksListIsEmptyError(Exception):
    pass


class IsNotDoneYetError(Exception):
    pass


class NormalOnDoneHandler:
    def __init__(self, normal_info: 'NormalInfo') -> None:
        self.normal_info: 'NormalInfo' = normal_info
    
    def __call__(self, task: Task) -> Any:
        self.normal_info.add_result(task.coro_id, task._result, task._exception)


class NormalInfo:
    def __init__(self, wait_done_id: str, tasks: List[Task]) -> None:
        self.tasks: List[Task] = tasks
        self.coroutines_num: int = len(tasks)
        self.results: List[Tuple[CoroID, Any, Exception]] = list()
        self.wait_done_id: str = wait_done_id
        self.ignore: bool = False
        self._done: bool = False
        self._timeout: bool = False
        self._failure: bool = False
    
    def add_result(self, coro_id: CoroID, result: Any, exception: Exception) -> None:
        if self.ignore:
            return
        
        self.results.append((coro_id, result, exception))
        if len(self.results) >= self.coroutines_num:
            self.done = True
    
    @property
    def failure(self) -> bool:
        return self._failure
    
    @failure.setter
    def failure(self, value: bool) -> None:
        self._failure = value
    
    @property
    def done(self) -> bool:
        return self._done
    
    @done.setter
    def done(self, value: bool) -> None:
        if self._done:
            return
        
        self._done = value
        if value:
            self.ignore = True
            put_request_to_service(AsyncEventBus, self.wait_done_id, None, CoroPriority.high)
    
    @property
    def timeout(self) -> bool:
        return self._timeout
    
    def on_timeout(self) -> None:
        self._timeout = True
        self.done = True
    
    def gather(self) -> List[Tuple[CoroID, Any, Exception]]:
        if not self.done:
            raise IsNotDoneYetError
        
        return self.results


# =================================================================================


class FastestOnDoneHandler:
    def __init__(self, fastest_info: 'FastestInfo') -> None:
        self.fastest_info: 'FastestInfo' = fastest_info
    
    def __call__(self, task: Task) -> Any:
        self.fastest_info.add_result((task.coro_id, task._result, task._exception))


class FastestInfo(NormalInfo):
    def __init__(self, wait_done_id: str, tasks: List[Task]) -> None:
        super().__init__(wait_done_id, tasks)
        self.coroutines_num: int = 1


# =================================================================================


class FastestSuccessfulOnDoneHandler:
    def __init__(self, fastest_successful_info: 'FastestSuccessfulInfo') -> None:
        self.fastest_successful_info: 'FastestSuccessfulInfo' = fastest_successful_info
    
    def __call__(self, task: Task) -> Any:
        self.fastest_successful_info.add_result((task.coro_id, task._result, task._exception))


class FastestSuccessfulInfo(NormalInfo):
    def __init__(self, wait_done_id: str, tasks: List[Task]) -> None:
        super().__init__(wait_done_id, tasks)
        self.fastest_successful_result: Optional[Tuple[CoroID, Any, Exception]] = None
    
    def add_result(self, coro_id: CoroID, result: Any, exception: Exception) -> None:
        if self.ignore:
            return
        
        self.results.append((coro_id, result, exception))
        if exception is None:
            self.fastest_successful_result = (coro_id, result, exception)
            self.done = True
        elif (len(self.results) >= self.coroutines_num):
            self.done = True
    
    def gather(self) -> List[Tuple[CoroID, Any, Exception]]:
        if not self.done:
            raise IsNotDoneYetError
        
        if self.fastest_successful_result is None:
            self.failure = True
        
        return [self.fastest_successful_result]


# =================================================================================


class FastestExceptionOnDoneHandler:
    def __init__(self, fastest_exception_info: 'FastestExceptionInfo') -> None:
        self.fastest_exception_info: 'FastestExceptionInfo' = fastest_exception_info
    
    def __call__(self, task: Task) -> Any:
        self.fastest_exception_info.add_result((task.coro_id, task._result, task._exception))


class FastestExceptionInfo(NormalInfo):
    def __init__(self, wait_done_id: str, tasks: List[Task]) -> None:
        super().__init__(wait_done_id, tasks)
        self.fastest_exception_result: Optional[Tuple[CoroID, Any, Exception]] = None
    
    def add_result(self, coro_id: CoroID, result: Any, exception: Exception) -> None:
        if self.ignore:
            return
        
        self.results.append((coro_id, result, exception))
        if exception is not None:
            self.fastest_exception_result = (coro_id, result, exception)
            self.done = True
        elif (len(self.results) >= self.coroutines_num):
            self.done = True
    
    def gather(self) -> List[Tuple[CoroID, Any, Exception]]:
        if not self.done:
            raise IsNotDoneYetError
        
        if self.fastest_exception_result is None:
            self.failure = True
        
        return [self.fastest_exception_result]


# =================================================================================


class FastestCustomOnDoneHandler:
    def __init__(self, fastest_custom_info: 'FastestCustomInfo') -> None:
        self.fastest_custom_info: 'FastestCustomInfo' = fastest_custom_info
    
    def __call__(self, task: Task) -> Any:
        self.fastest_custom_info.add_result((task.coro_id, task._result, task._exception))


class FastestCustomInfo(NormalInfo):
    def __init__(self, wait_done_id: str, tasks: List[Task], result_criteria_handler: Callable[[CoroID, Any, Exception], bool]) -> None:
        super().__init__(wait_done_id, tasks)
        self.result_criteria_handler: Callable[[CoroID, Any, Exception], bool] = result_criteria_handler
        self.fastest_custom_result: Optional[Tuple[CoroID, Any, Exception]] = None
    
    def add_result(self, coro_id: CoroID, result: Any, exception: Exception) -> None:
        if self.ignore:
            return
        
        self.results.append((coro_id, result, exception))
        if self.result_criteria_handler(coro_id, result, exception):
            self.fastest_custom_result = (coro_id, result, exception)
            self.done = True
        elif (len(self.results) >= self.coroutines_num):
            self.done = True
    
    def gather(self) -> List[Tuple[CoroID, Any, Exception]]:
        if not self.done:
            raise IsNotDoneYetError
        
        if self.fastest_custom_result is None:
            self.failure = True
        
        return [self.fastest_custom_result]


# =================================================================================


class AtomicOnDoneHandler:
    def __init__(self, atomic_info: 'AtomicInfo') -> None:
        self.atomic_info: 'AtomicInfo' = atomic_info
    
    def __call__(self, task: Task) -> Any:
        self.atomic_info.add_result(task.coro_id, task._result, task._exception)


class AtomicInfo(NormalInfo):
    def __init__(self, wait_done_id: str, tasks: List[Task], include_first_failure: bool = True) -> None:
        super().__init__(wait_done_id, tasks)
        self.include_first_failure: bool = include_first_failure
    
    def add_result(self, coro_id: CoroID, result: Any, exception: Exception) -> None:
        if self.ignore:
            return
        
        if exception is not None:
            if self.include_first_failure:
                self.results.append((coro_id, result, exception))
            
            self.failure = True
            self.done = True
        else:
            self.results.append((coro_id, result, exception))
        
        if len(self.results) >= self.coroutines_num:
            self.done = True
    
    def gather(self) -> List[Tuple[CoroID, Any, Exception]]:
        if not self.done:
            raise IsNotDoneYetError
        
        normalized_results: List[Tuple[CoroID, Any, Exception]] = list()
        provided_results: Dict[CoroID, Tuple[Any, Exception]] = dict()
        for result_info in self.results:
            if result_info is None:
                continue

            coro_id, result, exception = result_info
            provided_results[coro_id] = (result, exception)
        
        for task in self.tasks:
            coro_id = task.coro_id
            if coro_id in provided_results:
                result, exception = provided_results[coro_id]
                normalized_results.append((coro_id, result, exception))
            else:
                normalized_results.append(None)
        
        return normalized_results


# =================================================================================


class AtomicCustomOnDoneHandler:
    def __init__(self, atomic_custom_info: 'AtomicCustomInfo') -> None:
        self.atomic_custom_info: 'AtomicCustomInfo' = atomic_custom_info
    
    def __call__(self, task: Task) -> Any:
        self.atomic_custom_info.add_result(task.coro_id, task._result, task._exception)


def success_result_criteria_handler(coro_id: CoroID, result: Any, exception: Exception) -> bool:
    return exception is None


def exception_result_criteria_handler(coro_id: CoroID, result: Any, exception: Exception) -> bool:
    return exception is not None


def same_first_failed_result_formatter(coro_id: CoroID, result: Any, exception: Exception) -> Any:
    return (coro_id, result, exception)


def none_first_failed_result_formatter(coro_id: CoroID, result: Any, exception: Exception) -> Any:
    return None


class AtomicCustomInfo(NormalInfo):
    def __init__(
            self, 
            wait_done_id: str, 
            tasks: List[Task], 
            result_criteria_handler: Callable[[CoroID, Any, Exception], bool], 
            first_failed_result_formatter: Optional[Callable[[CoroID, Any, Exception], Any]] = None,
            ) -> None:
        super().__init__(wait_done_id, tasks)
        self.result_criteria_handler: Callable[[CoroID, Any, Exception], bool] = result_criteria_handler
        self.first_failed_result_formatter: Callable[[CoroID, Any, Exception], Any] = \
            none_first_failed_result_formatter if first_failed_result_formatter is None else first_failed_result_formatter
    
    def add_result(self, coro_id: CoroID, result: Any, exception: Exception) -> None:
        if self.ignore:
            return
        
        self.results.append((coro_id, result, exception))
        if (len(self.results) >= self.coroutines_num) or (exception is not None):
            self.done = True
        
        if not self.result_criteria_handler(coro_id, result, exception):
            result = self.first_failed_result_formatter(coro_id, result, exception)
            self.failure = True
            self.done = True
        else:
            result = (coro_id, result, exception)
        
        self.results.append(result)
        if len(self.results) >= self.coroutines_num:
            self.done = True
    
    def gather(self) -> List[Tuple[CoroID, Any, Exception]]:
        if not self.done:
            raise IsNotDoneYetError
        
        normalized_results: List[Tuple[CoroID, Any, Exception]] = list()
        provided_results: Dict[CoroID, Tuple[Any, Exception]] = dict()
        for result_info in self.results:
            if result_info is None:
                continue

            coro_id, result, exception = result_info
            provided_results[coro_id] = (result, exception)
        
        for task in self.tasks:
            coro_id = task.coro_id
            if coro_id in provided_results:
                result, exception = provided_results[coro_id]
                normalized_results.append((coro_id, result, exception))
            else:
                normalized_results.append(None)
        
        return normalized_results


# =================================================================================


class GracefulTerminationSettings:
    def __init__(self, 
                 phase_time_limit: Optional[float], 
                 ex_type: Type[Exception] = None, ex_value: Exception = None, ex_traceback: Any = None, 
                 first_phase_is_wait: bool = True,
                 last_phase_is_kill: bool = True, 
                 ) -> None:
        self.phase_time_limit: Optional[float] = phase_time_limit
        self.ex_type: Type[Exception] = ex_type
        self.ex_value: Exception = ex_value
        self.ex_traceback: Any = ex_traceback
        self.first_phase_is_wait: bool = first_phase_is_wait
        self.last_phase_is_kill: bool = last_phase_is_kill


class TerminationReason(IntEnum):
    success = 0
    failure = 1
    timeout = 2

                
async def atask_graceful_destroyer(
        i: Interface, 
        task: Task, 
        tree: bool = False,
        graceful_termination_settings: Optional[GracefulTerminationSettings] = None,
        ) -> None:
    await agraceful_coro_destroyer(
        i, 
        graceful_termination_settings.phase_time_limit, 
        task.coro_id, 
        graceful_termination_settings.ex_type, 
        graceful_termination_settings.ex_value, 
        graceful_termination_settings.ex_traceback, 
        tree, 
        graceful_termination_settings.first_phase_is_wait, 
        graceful_termination_settings.last_phase_is_kill, 
        )


async def aterminate_tasks_explicit(
        i: Interface, 
        tasks: List[Task], 
        termination_reason: TerminationReason,
        tree: bool = False,
        graceful_termination_settings: Optional[GracefulTerminationSettings] = None,
        terminate_other_tasks: Optional[bool] = None, 
        terminate_other_tasks_on_success: bool = False, 
        terminate_other_tasks_on_failure: bool = False, 
        terminate_other_tasks_on_timeout: bool = False, 
        wait_for_termination: bool = False, 
        ) -> bool:
    if not tasks:
        return False
    
    if terminate_other_tasks is not None:
        terminate_other_tasks_on_success = terminate_other_tasks
        terminate_other_tasks_on_failure = terminate_other_tasks
        terminate_other_tasks_on_timeout = terminate_other_tasks

    if ((TerminationReason.success == termination_reason) and terminate_other_tasks_on_success) or \
        ((TerminationReason.failure == termination_reason) and terminate_other_tasks_on_failure) or \
            ((TerminationReason.timeout == termination_reason) and terminate_other_tasks_on_timeout):
        if graceful_termination_settings is None:
            termination_tasks_params: List[KillSingleCoroParams] = list([KillSingleCoroParams(task.coro_id, tree) for task in tasks if not task.done])
            if not termination_tasks_params:
                return False
            
            i(KillCoroList, termination_tasks)
        else:
            termination_tasks_params: List[PSCP] = [PSCP(atask_graceful_destroyer, task, tree, graceful_termination_settings) for task in tasks if not task.done]
            if not termination_tasks_params:
                return False
            
            termination_tasks: List[Task] = await i(PutCoroList, termination_tasks_params, True)
            if wait_for_termination:
                await await_tasks_explicit(i, termination_tasks, False, None, None, False, False, False, False)
    
    return True


async def aterminate_tasks_implicit(
        tasks: List[Task], 
        termination_reason: TerminationReason,
        tree: bool = False,
        graceful_termination_settings: Optional[GracefulTerminationSettings] = None,
        terminate_other_tasks: Optional[bool] = None, 
        terminate_other_tasks_on_success: bool = False, 
        terminate_other_tasks_on_failure: bool = False, 
        terminate_other_tasks_on_timeout: bool = False, 
        wait_for_termination: bool = False, 
        ) -> bool:
    i: Interface = current_interface()
    return await aterminate_tasks_explicit(
        i, 
        tasks, 
        termination_reason, 
        tree, 
        graceful_termination_settings, 
        terminate_other_tasks, 
        terminate_other_tasks_on_success, 
        terminate_other_tasks_on_failure, 
        terminate_other_tasks_on_timeout, 
        wait_for_termination, 
        )


aterminate_tasks: Callable = aterminate_tasks_explicit
aterminate_tasks_im: Callable = aterminate_tasks_implicit


def terminate_tasks_explicit(
        i: Interface, 
        tasks: List[Task], 
        termination_reason: TerminationReason,
        tree: bool = False,
        graceful_termination_settings: Optional[GracefulTerminationSettings] = None,
        terminate_other_tasks: Optional[bool] = None, 
        terminate_other_tasks_on_success: bool = False, 
        terminate_other_tasks_on_failure: bool = False, 
        terminate_other_tasks_on_timeout: bool = False, 
        wait_for_termination: bool = False, 
        ) -> bool:
    i(RunCoro, aterminate_tasks_explicit,
        tasks, 
        termination_reason, 
        tree, 
        graceful_termination_settings, 
        terminate_other_tasks, 
        terminate_other_tasks_on_success, 
        terminate_other_tasks_on_failure, 
        terminate_other_tasks_on_timeout, 
        wait_for_termination, 
        )


def terminate_tasks_implicit(
        tasks: List[Task], 
        termination_reason: TerminationReason,
        tree: bool = False,
        graceful_termination_settings: Optional[GracefulTerminationSettings] = None,
        terminate_other_tasks: Optional[bool] = None, 
        terminate_other_tasks_on_success: bool = False, 
        terminate_other_tasks_on_failure: bool = False, 
        terminate_other_tasks_on_timeout: bool = False, 
        wait_for_termination: bool = False, 
        ) -> bool:
    i: Interface = current_interface()
    return terminate_tasks_explicit(
        i, 
        tasks, 
        termination_reason, 
        tree, 
        graceful_termination_settings, 
        terminate_other_tasks, 
        terminate_other_tasks_on_success, 
        terminate_other_tasks_on_failure, 
        terminate_other_tasks_on_timeout, 
        wait_for_termination, 
        )


terminate_tasks: Callable = terminate_tasks_explicit
terminate_tasks_im: Callable = terminate_tasks_implicit


# =================================================================================


async def await_tasks_explicit(
        i: Interface, 
        tasks: Union[Task, List[Task]], 
        terminate_tree: bool = False,
        graceful_termination_settings: Optional[GracefulTerminationSettings] = None,
        timeout: Optional[float] = None, 
        terminate_other_tasks: Optional[bool] = None, 
        terminate_other_tasks_on_success: bool = False, 
        terminate_other_tasks_on_failure: bool = False, 
        terminate_other_tasks_on_timeout: bool = False, 
        wait_for_termination: bool = False, 
        ) -> List[Tuple[CoroID, Any, Exception]]:
    if isinstance(tasks, Task):
        tasks = [tasks]
    
    if not tasks:
        raise TasksListIsEmptyError

    normal_wait_id: str = str(f'wait_task__normal__{uuid4()}')
    normal_info: NormalInfo = NormalInfo(normal_wait_id, tasks)

    if timeout is not None:
        # TODO: implement discard timer request when needed in order to improve performance
        timer_func_run_on(get_interface_and_loop_with_explicit_loop(i._loop), timeout, normal_info.on_timeout)

    for another_task in tasks:
        another_task.add_on_done_handler(NormalOnDoneHandler(normal_info))
    
    await i(AsyncEventBus, AsyncEventBusRequest().wait(normal_wait_id))
    results: List[Tuple[CoroID, Any, Exception]] = normal_info.gather()
    if normal_info.timeout:
        termination_reason: TerminationReason = TerminationReason.timeout
    elif normal_info.failure:
        termination_reason = TerminationReason.failure
    else:
        termination_reason = TerminationReason.success

    await terminate_tasks_explicit(
        i, 
        tasks, 
        termination_reason, 
        terminate_tree, 
        graceful_termination_settings, 
        terminate_other_tasks, 
        terminate_other_tasks_on_success, 
        terminate_other_tasks_on_failure, 
        terminate_other_tasks_on_timeout, 
        wait_for_termination, 
        )
    return results


async def await_tasks_implicit(
        tasks: Union[Task, List[Task]], 
        terminate_tree: bool = False,
        graceful_termination_settings: Optional[GracefulTerminationSettings] = None,
        timeout: Optional[float] = None, 
        terminate_other_tasks: Optional[bool] = None, 
        terminate_other_tasks_on_success: bool = False, 
        terminate_other_tasks_on_failure: bool = False, 
        terminate_other_tasks_on_timeout: bool = False, 
        wait_for_termination: bool = False, 
        ) -> List[Tuple[CoroID, Any, Exception]]:
    i: Interface = current_interface()
    return await await_tasks_explicit(
        i, 
        tasks, 
        terminate_tree, 
        graceful_termination_settings, 
        timeout, 
        terminate_other_tasks, 
        terminate_other_tasks_on_success, 
        terminate_other_tasks_on_failure, 
        terminate_other_tasks_on_timeout, 
        wait_for_termination, 
        )


await_tasks: Callable = await_tasks_explicit
await_tasks_im: Callable = await_tasks_implicit


def wait_tasks_explicit(
        i: Interface, 
        tasks: Union[Task, List[Task]], 
        terminate_tree: bool = False,
        graceful_termination_settings: Optional[GracefulTerminationSettings] = None,
        timeout: Optional[float] = None, 
        terminate_other_tasks: Optional[bool] = None, 
        terminate_other_tasks_on_success: bool = False, 
        terminate_other_tasks_on_failure: bool = False, 
        terminate_other_tasks_on_timeout: bool = False, 
        wait_for_termination: bool = False, 
        ) -> List[Tuple[CoroID, Any, Exception]]:
    i(RunCoro, await_tasks_explicit,
        tasks, 
        terminate_tree, 
        graceful_termination_settings, 
        timeout, 
        terminate_other_tasks, 
        terminate_other_tasks_on_success, 
        terminate_other_tasks_on_failure, 
        terminate_other_tasks_on_timeout, 
        wait_for_termination, 
        )


def wait_tasks_implicit(
        tasks: Union[Task, List[Task]], 
        terminate_tree: bool = False,
        graceful_termination_settings: Optional[GracefulTerminationSettings] = None,
        timeout: Optional[float] = None, 
        terminate_other_tasks: Optional[bool] = None, 
        terminate_other_tasks_on_success: bool = False, 
        terminate_other_tasks_on_failure: bool = False, 
        terminate_other_tasks_on_timeout: bool = False, 
        wait_for_termination: bool = False, 
        ) -> List[Tuple[CoroID, Any, Exception]]:
    i: Interface = current_interface()
    return wait_tasks_explicit(
        i, 
        tasks, 
        terminate_tree, 
        graceful_termination_settings, 
        timeout, 
        terminate_other_tasks, 
        terminate_other_tasks_on_success, 
        terminate_other_tasks_on_failure, 
        terminate_other_tasks_on_timeout, 
        wait_for_termination, 
        )


wait_tasks: Callable = wait_tasks_explicit
wait_tasks_im: Callable = wait_tasks_implicit


# =================================================================================


async def await_tasks_fastest_explicit(
        i: Interface, 
        tasks: Union[Task, List[Task]], 
        terminate_tree: bool = False,
        graceful_termination_settings: Optional[GracefulTerminationSettings] = None,
        timeout: Optional[float] = None, 
        terminate_other_tasks: Optional[bool] = None, 
        terminate_other_tasks_on_success: bool = False, 
        terminate_other_tasks_on_failure: bool = False, 
        terminate_other_tasks_on_timeout: bool = False, 
        wait_for_termination: bool = False, 
        ) -> Tuple[CoroID, Any]:
    if isinstance(tasks, Task):
        tasks = [tasks]
    
    if not tasks:
        raise TasksListIsEmptyError

    fastest_wait_id: str = str(f'wait_task__fastest__{uuid4()}')
    fastest_info: FastestInfo = FastestInfo(fastest_wait_id, tasks)

    if timeout is not None:
        # TODO: implement discard timer request when needed in order to improve performance
        timer_func_run_on(get_interface_and_loop_with_explicit_loop(i._loop), timeout, fastest_info.on_timeout)

    for another_task in tasks:
        another_task.add_on_done_handler(FastestOnDoneHandler(fastest_info))
    
    await i(AsyncEventBus, AsyncEventBusRequest().wait(fastest_wait_id))
    results: List[Tuple[CoroID, Any, Exception]] = fastest_info.gather()
    if fastest_info.timeout:
        termination_reason: TerminationReason = TerminationReason.timeout
    elif fastest_info.failure:
        termination_reason = TerminationReason.failure
    else:
        termination_reason = TerminationReason.success

    await aterminate_tasks_explicit(
        i, 
        tasks, 
        termination_reason, 
        terminate_tree, 
        graceful_termination_settings, 
        terminate_other_tasks, 
        terminate_other_tasks_on_success, 
        terminate_other_tasks_on_failure, 
        terminate_other_tasks_on_timeout, 
        wait_for_termination, 
        )
    if not results:
        return None, None
    
    coro_id, result, exception = results[0]
    if exception is not None:
        if isinstance(exception, TimeoutError) and (coro_id is not None):
            exception = SubTimeoutError()
        
        raise exception
    
    return coro_id, result


async def await_tasks_fastest_implicit(
        tasks: Union[Task, List[Task]], 
        terminate_tree: bool = False,
        graceful_termination_settings: Optional[GracefulTerminationSettings] = None,
        timeout: Optional[float] = None, 
        terminate_other_tasks: Optional[bool] = None, 
        terminate_other_tasks_on_success: bool = False, 
        terminate_other_tasks_on_failure: bool = False, 
        terminate_other_tasks_on_timeout: bool = False, 
        wait_for_termination: bool = False, 
        ) -> List[Tuple[CoroID, Any, Exception]]:
    i: Interface = current_interface()
    return await await_tasks_fastest_explicit(
        i, 
        tasks, 
        terminate_tree, 
        graceful_termination_settings, 
        timeout, 
        terminate_other_tasks, 
        terminate_other_tasks_on_success, 
        terminate_other_tasks_on_failure, 
        terminate_other_tasks_on_timeout, 
        wait_for_termination, 
        )


await_tasks_fastest: Callable = await_tasks_fastest_explicit
await_tasks_fastest_im: Callable = await_tasks_fastest_implicit


def wait_tasks_fastest_explicit(
        i: Interface, 
        tasks: Union[Task, List[Task]], 
        terminate_tree: bool = False,
        graceful_termination_settings: Optional[GracefulTerminationSettings] = None,
        timeout: Optional[float] = None, 
        terminate_other_tasks: Optional[bool] = None, 
        terminate_other_tasks_on_success: bool = False, 
        terminate_other_tasks_on_failure: bool = False, 
        terminate_other_tasks_on_timeout: bool = False, 
        wait_for_termination: bool = False, 
        ) -> List[Tuple[CoroID, Any, Exception]]:
    i(RunCoro, await_tasks_fastest_explicit,
        tasks, 
        terminate_tree, 
        graceful_termination_settings, 
        timeout, 
        terminate_other_tasks, 
        terminate_other_tasks_on_success, 
        terminate_other_tasks_on_failure, 
        terminate_other_tasks_on_timeout, 
        wait_for_termination, 
        )


def wait_tasks_fastest_implicit(
        tasks: Union[Task, List[Task]], 
        terminate_tree: bool = False,
        graceful_termination_settings: Optional[GracefulTerminationSettings] = None,
        timeout: Optional[float] = None, 
        terminate_other_tasks: Optional[bool] = None, 
        terminate_other_tasks_on_success: bool = False, 
        terminate_other_tasks_on_failure: bool = False, 
        terminate_other_tasks_on_timeout: bool = False, 
        wait_for_termination: bool = False, 
        ) -> List[Tuple[CoroID, Any, Exception]]:
    i: Interface = current_interface()
    return wait_tasks_fastest_explicit(
        i, 
        tasks, 
        terminate_tree, 
        graceful_termination_settings, 
        timeout, 
        terminate_other_tasks, 
        terminate_other_tasks_on_success, 
        terminate_other_tasks_on_failure, 
        terminate_other_tasks_on_timeout, 
        wait_for_termination, 
        )


wait_tasks_fastest: Callable = wait_tasks_fastest_explicit
wait_tasks_fastest_im: Callable = wait_tasks_fastest_implicit


# =================================================================================


async def await_tasks_fastest_successful_explicit(
        i: Interface, 
        tasks: Union[Task, List[Task]], 
        terminate_tree: bool = False,
        graceful_termination_settings: Optional[GracefulTerminationSettings] = None,
        timeout: Optional[float] = None, 
        terminate_other_tasks: Optional[bool] = None, 
        terminate_other_tasks_on_success: bool = False, 
        terminate_other_tasks_on_failure: bool = False, 
        terminate_other_tasks_on_timeout: bool = False, 
        wait_for_termination: bool = False, 
        ) -> Tuple[CoroID, Any]:
    if isinstance(tasks, Task):
        tasks = [tasks]
    
    if not tasks:
        raise TasksListIsEmptyError

    fastest_successful_wait_id: str = str(f'wait_task__fastest_successful__{uuid4()}')
    fastest_successful_info: FastestSuccessfulInfo = FastestSuccessfulInfo(fastest_successful_wait_id, tasks)

    if timeout is not None:
        # TODO: implement discard timer request when needed in order to improve performance
        timer_func_run_on(get_interface_and_loop_with_explicit_loop(i._loop), timeout, fastest_successful_info.on_timeout)

    for another_task in tasks:
        another_task.add_on_done_handler(FastestSuccessfulOnDoneHandler(fastest_successful_info))
    
    await i(AsyncEventBus, AsyncEventBusRequest().wait(fastest_successful_wait_id))
    results: List[Tuple[CoroID, Any, Exception]] = fastest_successful_info.gather()
    if fastest_successful_info.timeout:
        termination_reason: TerminationReason = TerminationReason.timeout
    elif fastest_successful_info.failure:
        termination_reason = TerminationReason.failure
    else:
        termination_reason = TerminationReason.success

    await aterminate_tasks_explicit(
        i, 
        tasks, 
        termination_reason, 
        terminate_tree, 
        graceful_termination_settings, 
        terminate_other_tasks, 
        terminate_other_tasks_on_success, 
        terminate_other_tasks_on_failure, 
        terminate_other_tasks_on_timeout, 
        wait_for_termination, 
        )
    if not results:
        return None, None
    
    coro_id, result, exception = results[0]
    if exception is not None:
        if isinstance(exception, TimeoutError) and (coro_id is not None):
            exception = SubTimeoutError()
        
        raise exception
    
    return coro_id, result


async def await_tasks_fastest_successful_implicit(
        tasks: Union[Task, List[Task]], 
        terminate_tree: bool = False,
        graceful_termination_settings: Optional[GracefulTerminationSettings] = None,
        timeout: Optional[float] = None, 
        terminate_other_tasks: Optional[bool] = None, 
        terminate_other_tasks_on_success: bool = False, 
        terminate_other_tasks_on_failure: bool = False, 
        terminate_other_tasks_on_timeout: bool = False, 
        wait_for_termination: bool = False, 
        ) -> List[Tuple[CoroID, Any, Exception]]:
    i: Interface = current_interface()
    return await await_tasks_fastest_successful_explicit(
        i, 
        tasks, 
        terminate_tree, 
        graceful_termination_settings, 
        timeout, 
        terminate_other_tasks, 
        terminate_other_tasks_on_success, 
        terminate_other_tasks_on_failure, 
        terminate_other_tasks_on_timeout, 
        wait_for_termination, 
        )


await_tasks_fastest_successful: Callable = await_tasks_fastest_successful_explicit
await_tasks_fastest_successful_im: Callable = await_tasks_fastest_successful_implicit


def wait_tasks_fastest_successful_explicit(
        i: Interface, 
        tasks: Union[Task, List[Task]], 
        terminate_tree: bool = False,
        graceful_termination_settings: Optional[GracefulTerminationSettings] = None,
        timeout: Optional[float] = None, 
        terminate_other_tasks: Optional[bool] = None, 
        terminate_other_tasks_on_success: bool = False, 
        terminate_other_tasks_on_failure: bool = False, 
        terminate_other_tasks_on_timeout: bool = False, 
        wait_for_termination: bool = False, 
        ) -> List[Tuple[CoroID, Any, Exception]]:
    i(RunCoro, await_tasks_explicit,
        tasks, 
        terminate_tree, 
        graceful_termination_settings, 
        timeout, 
        terminate_other_tasks, 
        terminate_other_tasks_on_success, 
        terminate_other_tasks_on_failure, 
        terminate_other_tasks_on_timeout, 
        wait_for_termination, 
        )


def wait_tasks_fastest_successful_implicit(
        tasks: Union[Task, List[Task]], 
        terminate_tree: bool = False,
        graceful_termination_settings: Optional[GracefulTerminationSettings] = None,
        timeout: Optional[float] = None, 
        terminate_other_tasks: Optional[bool] = None, 
        terminate_other_tasks_on_success: bool = False, 
        terminate_other_tasks_on_failure: bool = False, 
        terminate_other_tasks_on_timeout: bool = False, 
        wait_for_termination: bool = False, 
        ) -> List[Tuple[CoroID, Any, Exception]]:
    i: Interface = current_interface()
    return wait_tasks_fastest_successful_explicit(
        i, 
        tasks, 
        terminate_tree, 
        graceful_termination_settings, 
        timeout, 
        terminate_other_tasks, 
        terminate_other_tasks_on_success, 
        terminate_other_tasks_on_failure, 
        terminate_other_tasks_on_timeout, 
        wait_for_termination, 
        )


wait_tasks_fastest_successful: Callable = wait_tasks_fastest_successful_explicit
wait_tasks_fastest_successful_im: Callable = wait_tasks_fastest_successful_implicit


# =================================================================================


async def await_tasks_fastest_exception_explicit(
        i: Interface, 
        tasks: Union[Task, List[Task]], 
        terminate_tree: bool = False,
        graceful_termination_settings: Optional[GracefulTerminationSettings] = None,
        timeout: Optional[float] = None, 
        terminate_other_tasks: Optional[bool] = None, 
        terminate_other_tasks_on_success: bool = False, 
        terminate_other_tasks_on_failure: bool = False, 
        terminate_other_tasks_on_timeout: bool = False, 
        wait_for_termination: bool = False, 
        ) -> Tuple[CoroID, Exception]:
    if isinstance(tasks, Task):
        tasks = [tasks]
    
    if not tasks:
        raise TasksListIsEmptyError

    fastest_exception_wait_id: str = str(f'wait_task__fastest_exception__{uuid4()}')
    fastest_exception_info: FastestExceptionInfo = FastestExceptionInfo(fastest_exception_wait_id, tasks)

    if timeout is not None:
        # TODO: implement discard timer request when needed in order to improve performance
        timer_func_run_on(get_interface_and_loop_with_explicit_loop(i._loop), timeout, fastest_exception_info.on_timeout)

    for another_task in tasks:
        another_task.add_on_done_handler(FastestExceptionOnDoneHandler(fastest_exception_info))
    
    await i(AsyncEventBus, AsyncEventBusRequest().wait(fastest_exception_wait_id))
    results: List[Tuple[CoroID, Any, Exception]] = fastest_exception_info.gather()
    if fastest_exception_info.timeout:
        termination_reason: TerminationReason = TerminationReason.timeout
    elif fastest_exception_info.failure:
        termination_reason = TerminationReason.failure
    else:
        termination_reason = TerminationReason.success

    await aterminate_tasks_explicit(
        i, 
        tasks, 
        termination_reason, 
        terminate_tree, 
        graceful_termination_settings, 
        terminate_other_tasks, 
        terminate_other_tasks_on_success, 
        terminate_other_tasks_on_failure, 
        terminate_other_tasks_on_timeout, 
        wait_for_termination, 
        )
    if not results:
        return None, None
    
    coro_id, result, exception = results[0]
    if exception is not None:
        if isinstance(exception, TimeoutError) and (coro_id is not None):
            exception = SubTimeoutError()
    
    return coro_id, exception


async def await_tasks_fastest_exception_implicit(
        tasks: Union[Task, List[Task]], 
        terminate_tree: bool = False,
        graceful_termination_settings: Optional[GracefulTerminationSettings] = None,
        timeout: Optional[float] = None, 
        terminate_other_tasks: Optional[bool] = None, 
        terminate_other_tasks_on_success: bool = False, 
        terminate_other_tasks_on_failure: bool = False, 
        terminate_other_tasks_on_timeout: bool = False, 
        wait_for_termination: bool = False, 
        ) -> List[Tuple[CoroID, Any, Exception]]:
    i: Interface = current_interface()
    return await await_tasks_explicit(
        i, 
        tasks, 
        terminate_tree, 
        graceful_termination_settings, 
        timeout, 
        terminate_other_tasks, 
        terminate_other_tasks_on_success, 
        terminate_other_tasks_on_failure, 
        terminate_other_tasks_on_timeout, 
        wait_for_termination, 
        )


await_tasks_fastest_exception: Callable = await_tasks_fastest_exception_explicit
await_tasks_fastest_exception_im: Callable = await_tasks_fastest_exception_implicit


def wait_tasks_fastest_exception_explicit(
        i: Interface, 
        tasks: Union[Task, List[Task]], 
        terminate_tree: bool = False,
        graceful_termination_settings: Optional[GracefulTerminationSettings] = None,
        timeout: Optional[float] = None, 
        terminate_other_tasks: Optional[bool] = None, 
        terminate_other_tasks_on_success: bool = False, 
        terminate_other_tasks_on_failure: bool = False, 
        terminate_other_tasks_on_timeout: bool = False, 
        wait_for_termination: bool = False, 
        ) -> List[Tuple[CoroID, Any, Exception]]:
    i(RunCoro, await_tasks_fastest_exception_explicit,
        tasks, 
        terminate_tree, 
        graceful_termination_settings, 
        timeout, 
        terminate_other_tasks, 
        terminate_other_tasks_on_success, 
        terminate_other_tasks_on_failure, 
        terminate_other_tasks_on_timeout, 
        wait_for_termination, 
        )


def wait_tasks_fastest_exception_implicit(
        tasks: Union[Task, List[Task]], 
        terminate_tree: bool = False,
        graceful_termination_settings: Optional[GracefulTerminationSettings] = None,
        timeout: Optional[float] = None, 
        terminate_other_tasks: Optional[bool] = None, 
        terminate_other_tasks_on_success: bool = False, 
        terminate_other_tasks_on_failure: bool = False, 
        terminate_other_tasks_on_timeout: bool = False, 
        wait_for_termination: bool = False, 
        ) -> List[Tuple[CoroID, Any, Exception]]:
    i: Interface = current_interface()
    return wait_tasks_fastest_exception_explicit(
        i, 
        tasks, 
        terminate_tree, 
        graceful_termination_settings, 
        timeout, 
        terminate_other_tasks, 
        terminate_other_tasks_on_success, 
        terminate_other_tasks_on_failure, 
        terminate_other_tasks_on_timeout, 
        wait_for_termination, 
        )


wait_tasks_fastest_exception: Callable = wait_tasks_fastest_exception_explicit
wait_tasks_fastest_exception_im: Callable = wait_tasks_fastest_exception_implicit


# =================================================================================


async def await_tasks_fastest_custom_explicit(
        i: Interface, 
        tasks: Union[Task, List[Task]], 
        result_criteria_handler: Callable[[CoroID, Any, Exception], bool], 
        terminate_tree: bool = False,
        graceful_termination_settings: Optional[GracefulTerminationSettings] = None,
        timeout: Optional[float] = None, 
        terminate_other_tasks: Optional[bool] = None, 
        terminate_other_tasks_on_success: bool = False, 
        terminate_other_tasks_on_failure: bool = False, 
        terminate_other_tasks_on_timeout: bool = False, 
        wait_for_termination: bool = False, 
        ) -> Tuple[CoroID, Any]:
    if isinstance(tasks, Task):
        tasks = [tasks]
    
    if not tasks:
        raise TasksListIsEmptyError

    fastest_custom_wait_id: str = str(f'wait_task__fastest_custom__{uuid4()}')
    fastest_custom_info: FastestCustomInfo = FastestCustomInfo(fastest_custom_wait_id, tasks, result_criteria_handler)

    if timeout is not None:
        # TODO: implement discard timer request when needed in order to improve performance
        timer_func_run_on(get_interface_and_loop_with_explicit_loop(i._loop), timeout, fastest_custom_info.on_timeout)

    for another_task in tasks:
        another_task.add_on_done_handler(FastestCustomOnDoneHandler(fastest_custom_info))
    
    await i(AsyncEventBus, AsyncEventBusRequest().wait(fastest_custom_wait_id))
    results: List[Tuple[CoroID, Any, Exception]] = fastest_custom_info.gather()
    if fastest_custom_info.timeout:
        termination_reason: TerminationReason = TerminationReason.timeout
    elif fastest_custom_info.failure:
        termination_reason = TerminationReason.failure
    else:
        termination_reason = TerminationReason.success

    await aterminate_tasks_explicit(
        i, 
        tasks, 
        termination_reason, 
        terminate_tree, 
        graceful_termination_settings, 
        terminate_other_tasks, 
        terminate_other_tasks_on_success, 
        terminate_other_tasks_on_failure, 
        terminate_other_tasks_on_timeout, 
        wait_for_termination, 
        )
    if not results:
        return None, None
    
    coro_id, result, exception = results[0]
    if exception is not None:
        if isinstance(exception, TimeoutError) and (coro_id is not None):
            exception = SubTimeoutError()
        
        raise exception
    
    return coro_id, result


async def await_tasks_fastest_custom_implicit(
        tasks: Union[Task, List[Task]], 
        result_criteria_handler: Callable[[CoroID, Any, Exception], bool], 
        terminate_tree: bool = False,
        graceful_termination_settings: Optional[GracefulTerminationSettings] = None,
        timeout: Optional[float] = None, 
        terminate_other_tasks: Optional[bool] = None, 
        terminate_other_tasks_on_success: bool = False, 
        terminate_other_tasks_on_failure: bool = False, 
        terminate_other_tasks_on_timeout: bool = False, 
        wait_for_termination: bool = False, 
        ) -> List[Tuple[CoroID, Any, Exception]]:
    i: Interface = current_interface()
    return await await_tasks_fastest_custom_explicit(
        i, 
        tasks, 
        result_criteria_handler, 
        terminate_tree, 
        graceful_termination_settings, 
        timeout, 
        terminate_other_tasks, 
        terminate_other_tasks_on_success, 
        terminate_other_tasks_on_failure, 
        terminate_other_tasks_on_timeout, 
        wait_for_termination, 
        )


await_tasks_fastest_custom: Callable = await_tasks_fastest_custom_explicit
await_tasks_fastest_custom_im: Callable = await_tasks_fastest_custom_implicit


def wait_tasks_fastest_custom_explicit(
        i: Interface, 
        tasks: Union[Task, List[Task]], 
        result_criteria_handler: Callable[[CoroID, Any, Exception], bool], 
        terminate_tree: bool = False,
        graceful_termination_settings: Optional[GracefulTerminationSettings] = None,
        timeout: Optional[float] = None, 
        terminate_other_tasks: Optional[bool] = None, 
        terminate_other_tasks_on_success: bool = False, 
        terminate_other_tasks_on_failure: bool = False, 
        terminate_other_tasks_on_timeout: bool = False, 
        wait_for_termination: bool = False, 
        ) -> List[Tuple[CoroID, Any, Exception]]:
    i(RunCoro, await_tasks_fastest_custom_explicit,
        tasks, 
        result_criteria_handler, 
        terminate_tree, 
        graceful_termination_settings, 
        timeout, 
        terminate_other_tasks, 
        terminate_other_tasks_on_success, 
        terminate_other_tasks_on_failure, 
        terminate_other_tasks_on_timeout, 
        wait_for_termination, 
        )


def wait_tasks_fastest_custom_implicit(
        tasks: Union[Task, List[Task]], 
        result_criteria_handler: Callable[[CoroID, Any, Exception], bool], 
        terminate_tree: bool = False,
        graceful_termination_settings: Optional[GracefulTerminationSettings] = None,
        timeout: Optional[float] = None, 
        terminate_other_tasks: Optional[bool] = None, 
        terminate_other_tasks_on_success: bool = False, 
        terminate_other_tasks_on_failure: bool = False, 
        terminate_other_tasks_on_timeout: bool = False, 
        wait_for_termination: bool = False, 
        ) -> List[Tuple[CoroID, Any, Exception]]:
    i: Interface = current_interface()
    return wait_tasks_fastest_custom_explicit(
        i, 
        tasks, 
        result_criteria_handler, 
        terminate_tree, 
        graceful_termination_settings, 
        timeout, 
        terminate_other_tasks, 
        terminate_other_tasks_on_success, 
        terminate_other_tasks_on_failure, 
        terminate_other_tasks_on_timeout, 
        wait_for_termination, 
        )


wait_tasks_fastest_custom: Callable = wait_tasks_fastest_custom_explicit
wait_tasks_fastest_custom_im: Callable = wait_tasks_fastest_custom_implicit


# =================================================================================


async def await_tasks_atomic_explicit(
        i: Interface, 
        tasks: Union[Task, List[Task]], 
        terminate_tree: bool = False,
        graceful_termination_settings: Optional[GracefulTerminationSettings] = None,
        timeout: Optional[float] = None, 
        terminate_other_tasks: Optional[bool] = None, 
        terminate_other_tasks_on_success: bool = False, 
        terminate_other_tasks_on_failure: bool = False, 
        terminate_other_tasks_on_timeout: bool = False, 
        wait_for_termination: bool = False, 
        ):
    if terminate_other_tasks is not None:
        terminate_other_tasks_on_success = terminate_other_tasks
        terminate_other_tasks_on_failure = terminate_other_tasks
        terminate_other_tasks_on_timeout = terminate_other_tasks

    if isinstance(tasks, Task):
        tasks = [tasks]
    
    if not tasks:
        raise TasksListIsEmptyError

    atomic_wait_id: str = str(f'wait_task__atomic__{uuid4()}')
    atomic_info: AtomicInfo = AtomicInfo(atomic_wait_id, tasks)

    if timeout is not None:
        # TODO: implement discard timer request when needed in order to improve performance
        timer_func_run_on(get_interface_and_loop_with_explicit_loop(i._loop), timeout, atomic_info.on_timeout)

    for another_task in tasks:
        another_task.add_on_done_handler(AtomicOnDoneHandler(atomic_info))
    
    await i(AsyncEventBus, AsyncEventBusRequest().wait(atomic_wait_id))
    results: List[Tuple[CoroID, Any, Exception]] = atomic_info.gather()
    if atomic_info.timeout:
        termination_reason: TerminationReason = TerminationReason.timeout
    elif atomic_info.failure:
        termination_reason = TerminationReason.failure
    else:
        termination_reason = TerminationReason.success

    await aterminate_tasks_explicit(
        i, 
        tasks, 
        termination_reason, 
        terminate_tree, 
        graceful_termination_settings, 
        terminate_other_tasks, 
        terminate_other_tasks_on_success, 
        terminate_other_tasks_on_failure, 
        terminate_other_tasks_on_timeout, 
        wait_for_termination, 
        )
    return results


async def await_tasks_atomic_implicit(
        tasks: Union[Task, List[Task]], 
        terminate_tree: bool = False,
        graceful_termination_settings: Optional[GracefulTerminationSettings] = None,
        timeout: Optional[float] = None, 
        terminate_other_tasks: Optional[bool] = None, 
        terminate_other_tasks_on_success: bool = False, 
        terminate_other_tasks_on_failure: bool = False, 
        terminate_other_tasks_on_timeout: bool = False, 
        wait_for_termination: bool = False, 
        ) -> List[Tuple[CoroID, Any, Exception]]:
    i: Interface = current_interface()
    return await await_tasks_atomic_explicit(
        i, 
        tasks, 
        terminate_tree, 
        graceful_termination_settings, 
        timeout, 
        terminate_other_tasks, 
        terminate_other_tasks_on_success, 
        terminate_other_tasks_on_failure, 
        terminate_other_tasks_on_timeout, 
        wait_for_termination, 
        )


await_tasks_atomic: Callable = await_tasks_atomic_explicit
await_tasks_atomic_im: Callable = await_tasks_atomic_implicit


def wait_tasks_atomic_explicit(
        i: Interface, 
        tasks: Union[Task, List[Task]], 
        terminate_tree: bool = False,
        graceful_termination_settings: Optional[GracefulTerminationSettings] = None,
        timeout: Optional[float] = None, 
        terminate_other_tasks: Optional[bool] = None, 
        terminate_other_tasks_on_success: bool = False, 
        terminate_other_tasks_on_failure: bool = False, 
        terminate_other_tasks_on_timeout: bool = False, 
        wait_for_termination: bool = False, 
        ) -> List[Tuple[CoroID, Any, Exception]]:
    i(RunCoro, await_tasks_atomic_explicit,
        tasks, 
        terminate_tree, 
        graceful_termination_settings, 
        timeout, 
        terminate_other_tasks, 
        terminate_other_tasks_on_success, 
        terminate_other_tasks_on_failure, 
        terminate_other_tasks_on_timeout, 
        wait_for_termination, 
        )


def wait_tasks_atomic_implicit(
        tasks: Union[Task, List[Task]], 
        terminate_tree: bool = False,
        graceful_termination_settings: Optional[GracefulTerminationSettings] = None,
        timeout: Optional[float] = None, 
        terminate_other_tasks: Optional[bool] = None, 
        terminate_other_tasks_on_success: bool = False, 
        terminate_other_tasks_on_failure: bool = False, 
        terminate_other_tasks_on_timeout: bool = False, 
        wait_for_termination: bool = False, 
        ) -> List[Tuple[CoroID, Any, Exception]]:
    i: Interface = current_interface()
    return wait_tasks_atomic_explicit(
        i, 
        tasks, 
        terminate_tree, 
        graceful_termination_settings, 
        timeout, 
        terminate_other_tasks, 
        terminate_other_tasks_on_success, 
        terminate_other_tasks_on_failure, 
        terminate_other_tasks_on_timeout, 
        wait_for_termination, 
        )


wait_tasks_atomic: Callable = wait_tasks_atomic_explicit
wait_tasks_atomic_im: Callable = wait_tasks_atomic_implicit


# =================================================================================


async def await_tasks_atomic_custom_explicit(
        i: Interface, 
        tasks: Union[Task, List[Task]], 
        result_criteria_handler: Callable[[CoroID, Any, Exception], bool], 
        first_failed_result_formatter: Optional[Callable[[CoroID, Any, Exception], Any]] = None, 
        terminate_tree: bool = False,
        graceful_termination_settings: Optional[GracefulTerminationSettings] = None,
        timeout: Optional[float] = None, 
        terminate_other_tasks: Optional[bool] = None, 
        terminate_other_tasks_on_success: bool = False, 
        terminate_other_tasks_on_failure: bool = False, 
        terminate_other_tasks_on_timeout: bool = False, 
        wait_for_termination: bool = False, 
        ):
    if terminate_other_tasks is not None:
        terminate_other_tasks_on_success = terminate_other_tasks
        terminate_other_tasks_on_failure = terminate_other_tasks
        terminate_other_tasks_on_timeout = terminate_other_tasks

    if isinstance(tasks, Task):
        tasks = [tasks]
    
    if not tasks:
        raise TasksListIsEmptyError

    atomic_custom_wait_id: str = str(f'wait_task__atomic_custom__{uuid4()}')
    atomic_custom_info: AtomicCustomInfo = AtomicCustomInfo(
        atomic_custom_wait_id, 
        tasks, 
        result_criteria_handler, 
        first_failed_result_formatter, 
        )

    if timeout is not None:
        # TODO: implement discard timer request when needed in order to improve performance
        timer_func_run_on(get_interface_and_loop_with_explicit_loop(i._loop), timeout, atomic_custom_info.on_timeout)

    for another_task in tasks:
        another_task.add_on_done_handler(AtomicCustomOnDoneHandler(atomic_custom_info))
    
    await i(AsyncEventBus, AsyncEventBusRequest().wait(atomic_custom_wait_id))
    results: List[Tuple[CoroID, Any, Exception]] = atomic_custom_info.gather()
    if atomic_custom_info.timeout:
        termination_reason: TerminationReason = TerminationReason.timeout
    elif atomic_custom_info.failure:
        termination_reason = TerminationReason.failure
    else:
        termination_reason = TerminationReason.success

    await aterminate_tasks_explicit(
        i, 
        tasks, 
        termination_reason, 
        terminate_tree, 
        graceful_termination_settings, 
        terminate_other_tasks, 
        terminate_other_tasks_on_success, 
        terminate_other_tasks_on_failure, 
        terminate_other_tasks_on_timeout, 
        wait_for_termination, 
        )
    return results


async def await_tasks_atomic_custom_implicit(
        tasks: Union[Task, List[Task]], 
        result_criteria_handler: Callable[[CoroID, Any, Exception], bool], 
        first_failed_result_formatter: Optional[Callable[[CoroID, Any, Exception], Any]] = None, 
        terminate_tree: bool = False,
        graceful_termination_settings: Optional[GracefulTerminationSettings] = None,
        timeout: Optional[float] = None, 
        terminate_other_tasks: Optional[bool] = None, 
        terminate_other_tasks_on_success: bool = False, 
        terminate_other_tasks_on_failure: bool = False, 
        terminate_other_tasks_on_timeout: bool = False, 
        wait_for_termination: bool = False, 
        ) -> List[Tuple[CoroID, Any, Exception]]:
    i: Interface = current_interface()
    return await await_tasks_atomic_custom_explicit(
        i, 
        tasks, 
        result_criteria_handler, 
        first_failed_result_formatter, 
        terminate_tree, 
        graceful_termination_settings, 
        timeout, 
        terminate_other_tasks, 
        terminate_other_tasks_on_success, 
        terminate_other_tasks_on_failure, 
        terminate_other_tasks_on_timeout, 
        wait_for_termination, 
        )


await_tasks_atomic_custom: Callable = await_tasks_atomic_custom_explicit
await_tasks_atomic_custom_im: Callable = await_tasks_atomic_custom_implicit


def wait_tasks_atomic_custom_explicit(
        i: Interface, 
        tasks: Union[Task, List[Task]], 
        result_criteria_handler: Callable[[CoroID, Any, Exception], bool], 
        first_failed_result_formatter: Optional[Callable[[CoroID, Any, Exception], Any]] = None, 
        terminate_tree: bool = False,
        graceful_termination_settings: Optional[GracefulTerminationSettings] = None,
        timeout: Optional[float] = None, 
        terminate_other_tasks: Optional[bool] = None, 
        terminate_other_tasks_on_success: bool = False, 
        terminate_other_tasks_on_failure: bool = False, 
        terminate_other_tasks_on_timeout: bool = False, 
        wait_for_termination: bool = False, 
        ) -> List[Tuple[CoroID, Any, Exception]]:
    i(RunCoro, await_tasks_atomic_custom_explicit,
        tasks, 
        result_criteria_handler, 
        first_failed_result_formatter, 
        terminate_tree, 
        graceful_termination_settings, 
        timeout, 
        terminate_other_tasks, 
        terminate_other_tasks_on_success, 
        terminate_other_tasks_on_failure, 
        terminate_other_tasks_on_timeout, 
        wait_for_termination, 
        )


def wait_tasks_atomic_custom_implicit(
        tasks: Union[Task, List[Task]], 
        result_criteria_handler: Callable[[CoroID, Any, Exception], bool], 
        first_failed_result_formatter: Optional[Callable[[CoroID, Any, Exception], Any]] = None, 
        terminate_tree: bool = False,
        graceful_termination_settings: Optional[GracefulTerminationSettings] = None,
        timeout: Optional[float] = None, 
        terminate_other_tasks: Optional[bool] = None, 
        terminate_other_tasks_on_success: bool = False, 
        terminate_other_tasks_on_failure: bool = False, 
        terminate_other_tasks_on_timeout: bool = False, 
        wait_for_termination: bool = False, 
        ) -> List[Tuple[CoroID, Any, Exception]]:
    i: Interface = current_interface()
    return wait_tasks_atomic_custom_explicit(
        i, 
        tasks, 
        result_criteria_handler, 
        first_failed_result_formatter, 
        terminate_tree, 
        graceful_termination_settings, 
        timeout, 
        terminate_other_tasks, 
        terminate_other_tasks_on_success, 
        terminate_other_tasks_on_failure, 
        terminate_other_tasks_on_timeout, 
        wait_for_termination, 
        )


wait_tasks_atomic_custom: Callable = wait_tasks_atomic_custom_explicit
wait_tasks_atomic_custom_im: Callable = wait_tasks_atomic_custom_implicit


# =================================================================================


def wait_task_explicit(i: Interface, task: Union[Task, List[Task]]):
    wait_id: str = str(f'wait_task__{uuid4()}')
    def on_done_handler():
        put_request_to_service(AsyncEventBus, wait_id, None, CoroPriority.high)

    task.add_on_done_handler(on_done_handler)
    i(AsyncEventBus, AsyncEventBusRequest().wait(wait_id))
    return task.result


def wait_task_implicit(task: Union[Task, List[Task]]):
    i: Interface = current_interface()
    wait_id: str = str(f'wait_task__{uuid4()}')
    def on_done_handler():
        put_request_to_service(AsyncEventBus, wait_id, None, CoroPriority.high)

    task.add_on_done_handler(on_done_handler)
    i(AsyncEventBus, AsyncEventBusRequest().wait(wait_id))
    return task.result


wait_task = wait_task_explicit
wait_task_im = wait_task_implicit


async def await_task_explicit(i: Interface, task: Union[Task, List[Task]]):
    wait_id: str = str(f'wait_task__{uuid4()}')
    def on_done_handler():
        put_request_to_service(AsyncEventBus, wait_id, None, CoroPriority.high)

    task.add_on_done_handler(on_done_handler)
    await i(AsyncEventBus, AsyncEventBusRequest().wait(wait_id))
    return task.result


async def await_task_implicit(task: Union[Task, List[Task]]):
    i: Interface = current_interface()
    wait_id: str = str(f'wait_task__{uuid4()}')
    def on_done_handler():
        put_request_to_service(AsyncEventBus, wait_id, None, CoroPriority.high)

    task.add_on_done_handler(on_done_handler)
    await i(AsyncEventBus, AsyncEventBusRequest().wait(wait_id))
    return task.result


await_task = await_task_explicit
await_task_im = await_task_implicit
